import json
import subprocess
from pathlib import Path

from metadocker.common.errors import MetadockerError


class ExecutionError(MetadockerError):
    def __init__(self, cmdline, stdout, stderr):
        super().__init__(cmdline, stdout, stderr)
        self.cmdline = " ".join([str(x) for x in cmdline]) if not isinstance(cmdline, str) else cmdline
        self.stdout=stdout
        self.stderr = stderr

    def get_error(self):
        return f"ExecutionError: \nCommand Line: {self.cmdline}\nstdout: {self.stdout}\nstderr: {self.stderr}"

class DockerContainer:
    def __init__(self, data):
        self.name = data["Names"]
        self.image = data["Image"]
        self.id = data["ID"]
        self.status = data["State"]
        self.running = self.status == "running"

class DockerDriver:

    def __init__(self):
        pass

    def _run(self, *args, raise_on_fail=True, catch=True, **kwargs):

        if catch:
            kwargs.setdefault("stdout", subprocess.PIPE)
            kwargs.setdefault("stderr", subprocess.PIPE)

        print(f"Running: {' '.join(str(x) for x in args)}")
        ret = subprocess.run([str(x) for x in args], **kwargs)
        if ret.returncode != 0 and raise_on_fail:
            raise ExecutionError(args, ret.stdout, ret.stderr)
        return ret


    def _docker(self, *args, raise_on_fail=True, catch=True, **kwargs):
        return self._run("docker", *args, raise_on_fail=raise_on_fail, catch=catch, **kwargs)

    def build(self, file, tag=None, **kwarsg):
        tag = [] if tag is None else ["-t", tag]
        file = Path(file)
        name = file.name
        direct = file.parent
        self._docker("build", "-f", file, *tag, direct, catch=False, **kwarsg)

    def run(self, *args, **kwarsg):
        self._docker("run", *args, **kwarsg)


    def is_running(self, name):
        for data in self.ps(name=name):
            if data.running:
                return True
        return False


    def exec(self, name, commande=None, interactive=False, **kwarsg):
        args = []
        if interactive:
            args.append("-i")

        args.append(name)
        if commande:
            args.append(commande)
        self._docker("exec", *args, **kwarsg)

    def start(self, *args, **kwarsg):
        self._docker("start", *args, **kwarsg)


    def ps(self, classe=DockerContainer, **filters):
        cmdline = ["ps", "-a",  "--format", "json", "--no-trunc"]
        for k, v in filters.items():
            cmdline.extend(("--filter", f"{k}={v}"))

        ret = self._docker(*cmdline)
        content = [classe(json.loads(x)) for x in ret.stdout.split(b"\n") if x]
        return content


    def stop(self, name, time=None, signal=None, **kwargs):
        cmdlint = ["stop", name]
        if time is not None:
            cmdlint.extend(["-t", time])
        if signal is not None:
            cmdlint.extend(["-s", signal])
        return self._docker(*cmdlint, **kwargs)

    def kill(self, name, signal=None, **kwargs):
        cmdlint = ["kill", name]
        if signal is not None:
            cmdlint.extend(["-s", signal])
        return self._docker(*cmdlint, **kwargs)

    def rm(self, name, force=False, **kwargs):
        cmdlint = ["rm"]
        if force: cmdlint.append("--force")
        cmdlint.append(name)
        return self._docker(*cmdlint, **kwargs)


    def logs(self, name, details=False, follow=False, since=None, tail=None, timestamps=False, until=None, **kwargs):
        cmdlint = ["logs"]
        if details: cmdlint.append("--details")
        if follow: cmdlint.append("--follow")
        if since: cmdlint.extend(("--since", since))
        if tail: cmdlint.extend(("--tail", tail))
        if timestamps: cmdlint.append("--timestamps")
        if tail: until.extend(("--until", tail))
        cmdlint.append(name)
        return self._docker(*cmdlint, **kwargs)


driver = DockerDriver()