import argparse
import os
import re
from pathlib import Path

from metadocker.docker import DockerBuilder
from metadocker.env.base import Env
from metadocker.env.var import VarStr, VarInt, VarPath, VarDict
from metadocker.register import register_env



@register_env
class PythonServerEnv(Env):
    _name_ = "PythonServer"
    _alias_ = ["python"]
    _vars_ = [
        *Env._vars_,

        VarStr("PIP_INSTALL", None, help="La ou les dépendeances pip à installer", ask=True),
        VarStr("GIT_REPO",  None, help="L'adresse du dépot git", ask=True),
        VarStr("PYTHON_VERSION", "3.11", help="La version de python à utiliser"),
        VarStr("GIT_VERSION", help="La version (branche ou tag) à utiliser. Utiliser None pour rester sur le master", null=True),
        VarStr("USER", os.environ["USER"], help="Le nom d'utilisateur à utiliser. L'utilisateur sera crée. Utiliser None pour rester en root"),#
        VarInt("UID", os.getuid(), help="L'UID de l'utilisateur, uniquement si USER est défini"), #
        VarPath("BUILD_DIR", "/build", help="Le dossier au sein du container ou est buildé le projet"),
        VarStr("APT_PACKAGES", help="La liste des packages apt à installer (séparé par des espaces)"),
        VarStr("RUN_BEFORE_SETUP", help="Une commande a utiliser juste avant de lancer la commande d'installation du module python"),
        VarStr("RUN_AFTER_SETUP", help="Une commande a utiliser juste après avoir lancé la commande d'installation du module python"),
        VarStr("GIT_INSTALL_CMD", "pip install .", help="La commande à utiliser pour installer le package python"),
        VarStr("RUN_CMD", help="La commande à installer pour lancer l'application. Par défaut: python -m {PROJECT}"),
   ]

    PIP_INSTALL: str
    GIT_REPO: str
    PYTHON_VERSION: str
    GIT_VERSION: str
    USER: str
    UID: int
    BUILD_DIR: Path
    APT_PACKAGES: str
    RUN_BEFORE_SETUP: str
    RUN_AFTER_SETUP: str
    GIT_INSTALL_CMD: str
    RUN_CMD: str


    def _get_packages(self):
        data = self.APT_PACKAGES
        packages = []
        if self.GIT_REPO:
            packages.append("git")

        if isinstance(data, ((tuple, list, set))):
            packages.extend(data)
        elif isinstance(data, str):
            packages.extend(re.sub("\s\+", " ", data).split(" "))
        return packages


    def _get_envs(self):
        envs=[f'APP_DIR="{self.APP_DIR}"']
        if self.ENV:
            for k, v in self.ENV.items():
                envs.append(f'{k}=\"{v}\"')
        for k, v in self.env.items():
            envs.append(f'CFG_{k}=\"{v}\"')
        return "\\\n    ".join(envs)

    def generate_docker_file(self):
        self.docker.from_(f"python:{self.PYTHON_VERSION}-slim")

        self.docker.env(self._get_envs())

        if self.COPY:
            for k, v in self.COPY.items():
                self.docker.copy(k, v)

        if self.PORTS:
            self.docker.expose(*self.PORTS.values())

        packages = self._get_packages()
        if packages:
            self.docker.run(f'apt update -y && apt install -y', *packages)
        self.docker.run("mkdir -p", self.BUILD_DIR, self.APP_DIR)


        if self.USER:
            self.docker.run(f'adduser --disabled-password --gecos "" --home "/home/{self.USER}"  '
                            f'--shell "/sbin/nologin"  --uid "{self.UID}"  {self.USER}')

            self.docker.run("mkdir -p", f"/home/{self.USER}", " ".join(self.VOLUMES.values()))
            self.docker.run(f"chown -R {self.USER}:{self.USER}", self.BUILD_DIR, f"/home/{self.USER}",
                       " ".join(self.VOLUMES.values()), self.APP_DIR)

            if self.debug:
                self.docker.run(f"echo 'root:root' | chpasswd")
                self.docker.run(f"echo '{self.USER}:root' | chpasswd")


            self.docker.user(self.USER)
            self.docker.env("USER", self.USER)
            self.docker.env("PATH", f"/home/{self.USER}/.local/bin:$PATH")


        if self.VOLUMES:
            self.docker.volume(" ".join(self.VOLUMES.values()))

        self.docker.workdir(self.BUILD_DIR)

        self.docker.run("mkdir -p $HOME/.pip", "&&",
                   r"echo '[global]' > $HOME/.pip/pip.conf &&",
                   r"echo 'index-url = http://download.zope.org/simple' >> $HOME/.pip/pip.conf &&",
                   r"echo 'trusted-host = download.zope.org' >> $HOME/.pip/pip.conf &&",
                   r"echo 'extra-index-url= https://data.gautrais.eu/fanch/stable' >> $HOME/.pip/pip.conf")

        # install
        if self.RUN_BEFORE_SETUP:
            self.docker.run(self.RUN_BEFORE_SETUP)

        if self.PIP_INSTALL:
            self.docker.run(f"pip install {self.PIP_INSTALL}")

        if self.GIT_REPO:
            if self.GIT_VERSION:
                self.docker.run(f"git clone {self.GIT_REPO} . && git checkout {self.GIT_VERSION}")
            else:
                self.docker.run(f"git clone {self.GIT_REPO} . ")

            self.docker.run(self.GIT_INSTALL_CMD)

        if self.RUN_AFTER_SETUP:
            self.docker.run(self.RUN_AFTER_SETUP)


        self.docker.workdir(self.APP_DIR)

        if self.RUN_CMD:
            self.docker.cmd(self.RUN_CMD)
        else:
            self.docker.run(f"python -m {self.PROJECT} migrate")
            self.docker.cmd(f"python -m {self.PROJECT} run")


