from pathlib import Path

from metadocker.common.errors import MetadockerError

NoneType=None.__class__

class Empty: pass

EmptyValue = Empty()

def NotEmpty(x):
    if not x and x is not None:
        raise Var.ValidationException(None, f"doit être non vide")
    return x

def get_types(x):
    if x is None:
        return tuple()
    if not isinstance(x, (list, tuple, set)):
        return (x,)
    return x

class EmptyObject:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

class Var:

    class VarException(MetadockerError):
        def get_error(self):
            raise NotImplementedError()

    class RequiredException(VarException):
        def __init__(self, name):
            super().__init__(name)
            self.name = name

        def get_error(self):
            return f"La variable {self.name} est requise"

    class ByPassException(Exception):
        pass

    class ValidationException(VarException):
        def __init__(self, name, err):
            super().__init__(name, err)
            self.name = name
            self.err = err

        def get_error(self):
            return f"La variable {self.name} est invalide : {self.err}"

    class TypeException(VarException):
        def __init__(self, name, types, found):
            super().__init__(name, types)
            self.name = name

            if isinstance(types, type):
                types = [types]
            types = [x.__name__ for x in types]
            self.types = tuple(types)
            self.found = found

        def get_error(self):
            return f"La variable {self.name} doit être de type {'('+', '.join(self.types)+')'}. Type trouvé {self.found.__class__.__name__}"

    def __init__(self, name, default_value=EmptyValue, required=False, validate=None,
                 expected_types=None, help=None, is_meta=False, null=False, ask=False):
        self.name = name
        self.default_value = default_value
        self.required = required
        self._validate = validate if validate else lambda x: x
        self.expected_types = expected_types
        self.help = help
        self.null = null
        self.is_meta = is_meta
        self.ask = ask

    @property
    def has_default(self):
        return self.default_value is not EmptyValue

    def validate(self, obj):
        if not hasattr(obj, self.name):
            if self.default_value is EmptyValue:
                if self.required:
                    raise self.RequiredException(self.name)
                else:
                    raise self.ByPassException()
            return self.default_value

        data = getattr(obj, self.name)

        if self.expected_types:
            if data is None and self.null:
                pass
            elif not isinstance(data, self.expected_types):
                raise self.TypeException(self.name, self.expected_types, data)

            data = self._validate(data)

        return data

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    def _parse(self, type, value):
        if self.null and value == "None":
            return None
        return type(value)

    def ask_value(self):
        while True:
            print(self.help)
            prompt = [self.name]
            has_default = self.default_value is not EmptyValue

            if has_default:
                prompt.append(f" [{self.default_value}]")

            value = input("".join(prompt)+": ")

            if not has_default and not value:
                print(f"Merci de donner une réponse")
                continue

            try:
                parsed_value = None
                for x in self.expected_types or []:
                    try:
                        parsed_value = self._parse(x, value)
                        return self.validate(EmptyObject(**{self.name: parsed_value}))
                    except: pass
                else:
                    if self.null and not value:
                        return self.default_value
                    raise Var.RequiredException(self.name)

            except MetadockerError as err:
                if isinstance(err, self.ValidationException):
                    err.name = self.name
                print(err.get_error())


class VarStr(Var):
    def __init__(self, name, default_value=EmptyValue, required=False, validate=NotEmpty, help=None,
                 is_meta=False, expected_types=None, null=False, ask=False):
        super().__init__(name, default_value, required, validate, help=help,
                         expected_types=(str, *get_types(expected_types)), is_meta=is_meta,
                         null=null or default_value is None, ask=ask)

class VarPath(Var):
    def __init__(self, name, default_value=EmptyValue, required=False, validate=None, help=None,
                 is_meta=False, expected_types=None, null=False, ask=False):
        super().__init__(name, default_value, required, help=help, is_meta=is_meta,
                         validate=validate or (lambda x: Path(x)),
                         expected_types=(str, Path, *get_types(expected_types)), null=null, ask=ask)


class VarInt(Var):
    def __init__(self, name, default_value=EmptyValue, required=False, validate=None, help=None,
                 is_meta=False, expected_types=None, null=False, ask=False):
        super().__init__(name, default_value, required, validate, help=help, is_meta=is_meta,
                         expected_types=(int, *get_types(expected_types)), null=null, ask=ask)

class VarNumber(Var):
    def __init__(self, name, default_value=EmptyValue, required=False, validate=None, help=None,
                 is_meta=False, expected_types=None, null=False, ask=False):
        super().__init__(name, default_value, required, validate, help=help, is_meta=is_meta,
                         expected_types=(float, int, *get_types(expected_types)), null=null, ask=ask)

class VarDict(Var):
    def __init__(self, name, default_value=EmptyValue, required=False, validate=None, help=None,
                 is_meta=False, expected_types=None, null=False, ask=False):
        super().__init__(name, default_value, required, validate, help=help, is_meta=is_meta,
                         expected_types=(dict, *get_types(expected_types)), null=null, ask=ask)

